from ConfigParser import (
    ConfigParser,
    NoSectionError)
import errno
import os
from pwd import getpwnam
import re
from StringIO import StringIO
import subprocess
import sys

from Cheetah.Template import Template

import hookenv


CHARM_DIR = os.getcwd()
HOME = '/home/ubuntu'
PROJECT_DIR = os.path.join(HOME, 'juju-reports')
INI = 'production.ini'


def get_ini_path():
    return os.path.join(PROJECT_DIR, INI)


def get_ini():
    config = ConfigParser()
    config.read(get_ini_path())
    return config


def get_template_path(name):
    return os.path.join(PROJECT_DIR, 'resources/{}.template'.format(name))


def install_ini(mongo_url, gh_token, jujuci_url, token):
    config = ConfigParser()
    path = get_template_path(INI)
    if not os.path.exists(path):
        path = get_ini_path()
    config.read(path)
    config.set('app:main', 'mongo.url', mongo_url)
    config.set('app:main', 'gh_token', gh_token)
    config.set('app:main', 'jujuci.url', jujuci_url)
    config.set('app:main', 'charm_bundle_test_token', token)
    config.set('app:main', 'home', HOME)
    config.write(open(get_ini_path(), 'w'))
    return config


def configure_s3(config):
    template_path = get_template_path('juju-qa.s3cfg')
    try:
        template_cm = open(template_path)
    except IOError as e:
        if e.errno == errno.ENOENT:
            hookenv.log('Skipping s3 config due to missing template.')
            return
    with template_cm as template_file:
        template = template_file.read()
    config_path = os.path.join(PROJECT_DIR, 'juju-qa.s3cfg')
    s3config = template.format(access_key=config['aws-access-key'],
                               secret_key=config['aws-secret-key'])
    update_file(config_path, s3config)
    home_stat = os.stat(HOME)
    data_path = os.path.join(HOME, 'juju-qa-data')
    ensure_dir(data_path)
    os.chown(data_path, home_stat.st_uid, home_stat.st_gid)
    cbtest_path = os.path.join(HOME, 'juju-qa-data/cbtest')
    ensure_dir(cbtest_path)
    os.chown(cbtest_path, home_stat.st_uid, home_stat.st_gid)


def ensure_dir(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def get_configured_port(ini):
    return ini_get(ini, 'server:main', 'port')


def set_port(ini, open_port=True):
    port = get_configured_port(ini)
    if port is None:
        return
    if open_port:
        hookenv.open_port(port)
    else:
        hookenv.close_port(port)


def call_script(script_path, root=False):
    env = os.environ.copy()
    ini = get_ini_path()
    env['HOME'] = HOME
    env['INI'] = ini
    if root:
        args = []
    else:
        args = ['su', 'ubuntu', '-p', '-c']

    args.append(script_path)
    subprocess.check_call(args, env=env, cwd=PROJECT_DIR)
    return env


def stop():
    try:
        call_script('scripts/stop', root=True)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def start():
    call_script('scripts/run')


def open_secret(path, uid=None, gid=None):
    fd = os.open(path, os.O_WRONLY | os.O_TRUNC | os.O_CREAT, 0600)
    if uid is not None:
        os.fchown(fd, uid, gid)
    return os.fdopen(fd, 'w')


def configure_lp(ssh_key):
    if ssh_key == '':
        return
    with open_secret('/root/.ssh/lp_rsa') as key_file:
        key_file.write(ssh_key)
    pwd = getpwnam('ubuntu')
    ubuntu_path = os.path.join(HOME, '.ssh/lp_rsa')
    with open_secret(ubuntu_path, pwd.pw_uid, pwd.pw_gid) as key_file:
        key_file.write(ssh_key)
    # Login to LP
    call_bzr(['lp-login', 'juju-qa-bot'])


def call_bzr(args):
    args = ['bzr', '--no-aliases'] + args
    return subprocess.check_call(args)


def update_source(source_url, revno):
    # Make it easier to develop on a deploy
    call_bzr(['bind', '-d', PROJECT_DIR, source_url])

    # Grab the code
    revno = 'revno:%s' % revno

    bzr_cmd = ['pull', '--overwrite', source_url, '-r', revno, '-d',
               PROJECT_DIR, '--remember', '--local']
    call_bzr(bzr_cmd)

    # Restore the templates if they are missing/altered.
    call_bzr(['revert', os.path.join(PROJECT_DIR, 'resources')])


def install_production(dist_clean, develop):
    subprocess.check_call(['make', 'deploydeps'], cwd=PROJECT_DIR)
    if dist_clean:
        subprocess.check_call(['make', 'distclean'], cwd=PROJECT_DIR)
    target = 'install' if develop else 'prodinstall'
    subprocess.check_call(['make', target], cwd=PROJECT_DIR)
    subprocess.check_call(['apt-get', 'autoremove', '--yes'])


def install_cronjob(interval, error_email):
    t = Template(file='templates/crontab.tmpl')
    t.interval = interval
    if error_email is not None:
        t.error_email = 'MAILTO=%s' % error_email
    else:
        t.error_email = ''
    t.reports_home = PROJECT_DIR
    open('/etc/cron.d/ubuntu', 'w').write(str(t))


def ini_get(ini, section, key):
    try:
        return ini.get(section, key)
    except NoSectionError:
        return


def get_mongo_url():
    """Use the 'database' relation settings to build a mongodb url.

    If no mongodb units are present, return an empty string.
    """
    host_ports = []
    replset = None
    for relation_id in hookenv.relation_ids('database'):
        for unit in hookenv.related_units(relation_id):
            relation_data = hookenv.relation_get(rid=relation_id, unit=unit)
            # Consider configuration incomplete until port is supplied.
            if (relation_data is None or 'port' not in relation_data or
                'hostname' not in relation_data):
                continue
            if replset is None:
                replset = relation_data['replset']
            # All units should be members of the same replication set.
            elif replset != relation_data['replset']:
                raise AssertionError('DB instances are from different sets!')
            host_ports.append(
                '%(hostname)s:%(port)s' % relation_data)
    if len(host_ports) == 0:
        return ''
    return 'mongodb://%s/?replicaSet=%s' % (','.join(host_ports), replset)


def update_website(ini):
    relation_settings = {
        'port': get_configured_port(ini),
        'hostname': hookenv.unit_private_ip(),
    }
    for relation_id in hookenv.relation_ids('website'):
        hookenv.relation_set(relation_id, relation_settings)


def update_file(path, new_contents):
    """Write file only if the contents are changed."""
    try:
        file_context = open(path, 'r')
    except IOError as e:
        if e.errno == errno.ENOENT:
            old_contents = None
        else:
            raise
    else:
        with file_context as fileobj:
            old_contents = fileobj.read()
    if old_contents == new_contents:
        return
    with open(path, 'w') as fileobj:
        fileobj.write(new_contents)


def update_from_config():
    ini = get_ini()
    mongo_url = get_mongo_url()
    config = hookenv.config()
    set_port(ini, False)
    stop()
    configure_lp(config['lp-key'])
    if config['source'] == '':
        hookenv.log('Incomplete config: source')
        return
    update_source(config['source'], config['revno'])
    oauth_path = os.path.join(PROJECT_DIR, 'jujureports/lp_credentials.txt')
    update_file(oauth_path, config['lp-oauth'] + '\n')
    install_production(config['dist-clean'], config['develop-install'])
    if mongo_url == '':
        hookenv.log('No mongodb set up.')
        return
    ini = install_ini(mongo_url, config['gh-token'], config['jujuci-url'],
                      config['charm-bundle-test-token'])
    configure_s3(config)
    for key in ['lp-key', 'lp-oauth', 'gh-token']:
        if config[key] == '':
            hookenv.log('Not setting up cron because {} not set.'.format(key))
            break
    else:
        install_cronjob(config['cron-interval'], config['error-email'])
    start()
    set_port(ini, True)
    update_website(ini)
